"""
constants
"""

# Django
from django.utils.text import slugify

# AA RSS to Discord
from aa_rss_to_discord import __version__

VERBOSE_NAME = (
    "AA RSS To Discord - Alliance Auth "
    "module to post news from RSS feeds to your Discord"
)

verbose_name_slugified: str = slugify(VERBOSE_NAME, allow_unicode=True)
github_url: str = "https://github.com/ppfeufer/aa-rss-to-discord"
USER_AGENT = f"{verbose_name_slugified} v{__version__} {github_url}"



"""
Our Models
"""

# Third Party
from aadiscordbot.models import Channels

# Django
from django.db import models
from django.utils.translation import gettext as _

# AA RSS to Discord
from aa_rss_to_discord.managers import RssFeedsManager


class RssFeeds(models.Model):
    """
    Model :: RssFeeds
    """

    name = models.CharField(max_length=254, default="")
    url = models.CharField(max_length=254, default="")
    discord_channel = models.ForeignKey(
        Channels,
        related_name="+",
        null=True,
        blank=True,
        default=None,
        on_delete=models.SET_NULL,
    )
    enabled = models.BooleanField(default=True)

    objects = RssFeedsManager()

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("RSS Feed")
        verbose_name_plural = _("RSS Feeds")

    def __str__(self):
        return f'RSS Feed "{self.name}" for channel {self.discord_channel}'


class LastItem(models.Model):
    """
    Model :: LastItem
    """

    rss_feed = models.ForeignKey(
        RssFeeds,
        related_name="+",
        on_delete=models.CASCADE,
    )
    rss_item_time = models.CharField(max_length=254, blank=True, default="")
    rss_item_title = models.CharField(max_length=254, default="")
    rss_item_link = models.CharField(max_length=254, default="")
    rss_item_guid = models.CharField(max_length=254, blank=True, default="")

    class Meta:
        """
        Meta definitions
        """

        default_permissions = ()
        verbose_name = _("Last Item")
        verbose_name_plural = _("Last Items")

    def __str__(self):
        return f'RSS Entry "{self.rss_item_title}"'



"""
"RSS" cog for discordbot - https://github.com/pvyParts/allianceauth-discordbot
"""

# Standard Library
import logging

# Third Party
import discord
from aadiscordbot.cogs.utils.decorators import sender_is_admin
from aadiscordbot.models import Channels, Servers
from discord.ext import commands

# Django
from django.core.exceptions import ValidationError
from django.core.validators import URLValidator

# AA RSS to Discord
from aa_rss_to_discord.models import RssFeeds

logger = logging.getLogger(__name__)


class Rss(commands.Cog):
    """
    RSS management via Discord
    """

    def __init__(self, bot):
        self.bot = bot

    @commands.command(pass_context=True)
    @sender_is_admin()
    async def rss_add(self, ctx, rss_url: str = None, *, rss_name: str = None):
        """
        Adding a RSS/Atom feed to the current Discord channel
        """

        await ctx.trigger_typing()

        if rss_url is None or rss_name is None:
            return await ctx.send(
                "To add a RSS/Atom feed to this channel, "
                "please use the following syntax.\n\n"
                "```!rss_add rss_url rss_name```\n\n"
                "Both arguments are required."
            )

        channel_name = ctx.message.channel.name
        channel_id = ctx.message.channel.id
        server_name = ctx.guild.name
        server_id = ctx.guild.id

        validate_url = URLValidator()

        # Check if rss_url is a valid URL
        try:
            validate_url(rss_url)
        except ValidationError:
            return await ctx.send("The URL provided is not valid!")

        # Check if there is already a RSS/Atom feed with this URL for this channel
        if RssFeeds.objects.filter(
            url__iexact=rss_url, discord_channel_id__exact=channel_id
        ).exists():
            return await ctx.send(
                "A RSS/Atom feed with this URL already exists for this channel"
            )

        # Check if the current server and channel are already
        # known to the allianceauth-discordbot module
        if not Servers.objects.filter(server__exact=server_id).exists():
            Servers(server=server_id, name=server_name).save()

        if not Channels.objects.filter(
            channel__exact=channel_id, server_id__exact=server_id
        ).exists():
            Channels(channel=channel_id, name=channel_name, server_id=server_id).save()

        # Add the new RSS/Atom feed
        RssFeeds(
            url=rss_url, name=rss_name, discord_channel_id=channel_id, enabled=True
        ).save()

        return await ctx.send(f'RSS/Atom feed "{rss_name}" added to this channel')

    @commands.command(pass_context=True)
    @sender_is_admin()
    async def rss_list(self, ctx):
        """
        List all RSS/Atom feeds for the current Discord channel
        """

        await ctx.trigger_typing()

        channel_id = ctx.message.channel.id
        payload = "No RSS/Atom feeds have been registered for this channel."
        rss_feeds = RssFeeds.objects.filter(discord_channel_id__exact=channel_id)

        if rss_feeds.count() > 0:
            payload = (
                "The following RSS/Atom feeds are registered for this channel:\n\n```"
            )

            for rss_feed in rss_feeds:
                rss_status = "Enabled" if rss_feed.enabled is True else "Disabled"

                payload += (
                    f"ID: {rss_feed.pk}\n"
                    f"Name: {rss_feed.name}\n"
                    f"URL: {rss_feed.url}\n"
                    f"Status: {rss_status}\n\n"
                )

            payload += "```"

        return await ctx.send(payload)

    @commands.command(pass_context=True)
    @sender_is_admin()
    async def rss_delete(self, ctx, rss_feed_id: int):
        """
        Remove a RSS/Atom feed from the current Discord channel
        """

        await ctx.trigger_typing()

        channel_id = ctx.message.channel.id

        try:
            rss_feed = RssFeeds.objects.get(
                pk=rss_feed_id, discord_channel_id__exact=channel_id
            )

            rss_feed_name = rss_feed.name

            rss_feed.delete()

            payload = (
                f'RSS/Atom feed "{rss_feed_name}" has been removed from this channel.'
            )
        except RssFeeds.DoesNotExist:
            payload = "This RSS/Atom feed does not exist in this Discord channel."

        return await ctx.send(payload)

    @rss_delete.error
    async def rss_delete_error(self, ctx, error):
        """
        Delete error

        :param ctx:
        :type ctx:
        :param error:
        :type error:
        :return:
        :rtype:
        """

        if isinstance(
            error,
            (
                discord.ext.commands.CommandInvokeError,
                discord.ext.commands.errors.MissingRequiredArgument,
            ),
        ):
            await ctx.send(
                "You didn't provide a numeric value for the RSS/Atom ID you want to "
                "remove.\n\nExample:\n```!rss_delete 5```To remove RSS/Atom feed "
                "wth the ID 5.\nYou find a list of RSS/Atom feeds for this channel "
                "with `!rss_list`"
            )

    @commands.command(pass_context=True)
    @sender_is_admin()
    async def rss_enable(self, ctx, rss_feed_id: int):
        """
        Enable a disabled RSS/Atom feed for the current Discord channel
        """

        await ctx.trigger_typing()

        channel_id = ctx.message.channel.id

        try:
            rss_feed = RssFeeds.objects.get(
                pk=rss_feed_id, discord_channel_id__exact=channel_id
            )

            rss_feed.enabled = 1
            rss_feed.save()

            payload = (
                f'RSS/Atom feed "{rss_feed.name}" has been enabled for this channel.'
            )
        except RssFeeds.DoesNotExist:
            payload = "This RSS/Atom feed does not exist in this Discord channel."

        return await ctx.send(payload)

    @rss_enable.error
    async def rss_enable_error(self, ctx, error):
        """
        Enable error

        :param ctx:
        :type ctx:
        :param error:
        :type error:
        :return:
        :rtype:
        """

        if isinstance(
            error,
            (
                discord.ext.commands.CommandInvokeError,
                discord.ext.commands.errors.MissingRequiredArgument,
            ),
        ):
            await ctx.send(
                "You didn't provide a numeric value for the RSS/Atom ID you want to "
                "enable.\n\nExample:\n```!rss_enable 5```To enable RSS/Atom feed wth "
                "the ID 5.\nYou find a list of RSS/Atom feeds for this channel "
                "with `!rss_list`"
            )

    @commands.command(pass_context=True)
    @sender_is_admin()
    async def rss_disable(self, ctx, rss_feed_id: int):
        """
        Disable an enabled RSS/Atom feed for the current Discord channel
        """

        await ctx.trigger_typing()

        channel_id = ctx.message.channel.id

        try:
            rss_feed = RssFeeds.objects.get(
                pk=rss_feed_id, discord_channel_id__exact=channel_id
            )

            rss_feed.enabled = 0
            rss_feed.save()

            payload = (
                f'RSS/Atom feed "{rss_feed.name}" has been disabled for this channel.'
            )
        except RssFeeds.DoesNotExist:
            payload = "This RSS/Atom feed does not exist in this Discord channel."

        return await ctx.send(payload)

    @rss_disable.error
    async def rss_disable_error(self, ctx, error):
        """
        Disable error

        :param ctx:
        :type ctx:
        :param error:
        :type error:
        :return:
        :rtype:
        """

        if isinstance(
            error,
            (
                discord.ext.commands.CommandInvokeError,
                discord.ext.commands.errors.MissingRequiredArgument,
            ),
        ):
            await ctx.send(
                "You didn't provide a numeric value for the RSS/Atom ID you want to "
                "enable.\n\nExample:\n```!rss_disable 5```To disable RSS/Atom feed "
                "wth the ID 5.\nYou find a list of RSS/Atom feeds for this channel "
                "with `!rss_list`"
            )


def setup(bot):
    """
    Setup the cog
    :param bot:
    """

    bot.add_cog(Rss(bot))



"""
Django admin declarations
"""

# Django
from django.contrib import admin

# AA RSS to Discord
from aa_rss_to_discord.models import RssFeeds


@admin.register(RssFeeds)
class RssFeedsAdmin(admin.ModelAdmin):
    """
    RssFeedsAdmin
    """

    list_display = ("name", "enabled", "url", "discord_channel")
    ordering = ("name",)
    list_filter = ("enabled", "discord_channel")
    search_fields = ("name", "url", "discord_channel")



"""
AA RSS To Discord Tasks
"""

# Standard Library
import logging
import re

# Third Party
import feedparser
from celery import shared_task

# Django
from django.apps import apps

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA RSS to Discord
from aa_rss_to_discord import __title__
from aa_rss_to_discord.models import LastItem, RssFeeds

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


def remove_emoji(string):
    """
    Removing these dumb as fuck emojis from the title string.
    Like honestly, who in the hell needs that shit?

    :param string:
    :type string:
    :return:
    :rtype:
    """

    emoji_pattern = re.compile(
        pattern="["
        "\U0001F600-\U0001F64F"  # Emoticons
        "\U0001F300-\U0001F5FF"  # Symbols & pictographs
        "\U0001F680-\U0001F6FF"  # Transport & map symbols
        "\U0001F1E0-\U0001F1FF"  # Flags (iOS)
        "\U00002500-\U00002BEF"  # Chinese char
        "\U00002702-\U000027B0"
        "\U00002702-\U000027B0"
        "\U000024C2-\U0001F251"
        "\U0001f926-\U0001f937"
        "\U00010000-\U0010ffff"
        "\u2640-\u2642"
        "\u2600-\u2B55"
        "\u200d"
        "\u23cf"
        "\u23e9"
        "\u231a"
        "\ufe0f"  # Dingbats
        "\u3030"
        "]+",
        flags=re.UNICODE,
    )

    return emoji_pattern.sub(repl=r"", string=string)


@shared_task(**{"base": QueueOnce})
def fetch_rss() -> None:  # pylint: disable=too-many-statements, too-many-branches
    """
    Fetch RSS feeds and post to Discord

    :return:
    :rtype:
    """

    # pylint: disable=too-many-nested-blocks
    if apps.is_installed(app_name="aadiscordbot"):
        # Third Party
        from aadiscordbot.tasks import (  # pylint: disable=import-outside-toplevel
            send_message,
        )

        rss_feeds = RssFeeds.objects.select_enabled()

        if rss_feeds:
            for rss_feed in rss_feeds:
                logger.info(msg=f'Fetching RSS Feed "{rss_feed.name}"')

                feed = feedparser.parse(url_file_stream_or_string=rss_feed.url)

                feed_entry_title = "No title"
                feed_entry_link = None
                feed_entry_time = None
                feed_entry_guid = None
                has_last_item = False
                last_item = None
                post_entry = False

                try:
                    latest_entry = feed.entries[0]

                    feed_entry_title = remove_emoji(
                        string=latest_entry.get("title", "No title")
                    )
                    feed_entry_link = latest_entry.get("link", None)
                    feed_entry_time = latest_entry.get(
                        "published", latest_entry.updated
                    )
                    feed_entry_guid = latest_entry.get("id", None)
                except AttributeError as exc:
                    logger.debug(
                        msg=f'Malformed RSS feed item in feed "{rss_feed.name}". Error: {exc}'
                    )
                except IndexError as exc:
                    logger.debug(
                        msg=f'Could not index the RSS feed "{rss_feed.name}". Error: {exc}'
                    )
                else:
                    post_entry = True
                    has_last_item = True

                    try:
                        last_item = LastItem.objects.get(rss_feed=rss_feed)

                        if (
                            last_item
                            and last_item.rss_item_time == feed_entry_time
                            and last_item.rss_item_title == feed_entry_title
                            and last_item.rss_item_link == feed_entry_link
                            and last_item.rss_item_guid == feed_entry_guid
                        ):
                            logger.debug(
                                msg=(
                                    f'News item "{feed_entry_title}" for RSS Feed '
                                    f'"{rss_feed.name}" has already been posted to your Discord'
                                )
                            )
                            post_entry = False
                    except LastItem.DoesNotExist:
                        logger.debug(msg="This seems to be a completely new RSS feed.")

                        has_last_item = False

                logger.debug(
                    msg=(
                        "RSS Information gathered: "
                        f"post_entry => {post_entry}, "
                        f'feed_entry_link => "{feed_entry_link}", '
                        f'feed_entry_title => "{feed_entry_title}", '
                        f"feed_entry_time => {feed_entry_time}, "
                        f"feed_entry_guid => {feed_entry_guid}"
                    )
                )

                if (
                    post_entry is True
                    and feed_entry_link is not None
                    and feed_entry_guid is not None
                ):
                    logger.info(
                        msg=(
                            "New entry found, posting to Discord channel "
                            f"{rss_feed.discord_channel}"
                        )
                    )

                    if has_last_item is True:
                        # Update the last item ...
                        last_item.rss_item_time = feed_entry_time
                        last_item.rss_item_title = feed_entry_title
                        last_item.rss_item_link = feed_entry_link
                        last_item.rss_item_guid = feed_entry_guid
                        last_item.save()
                    else:
                        # Set the last item ...
                        LastItem(
                            rss_feed=rss_feed,
                            rss_item_time=feed_entry_time,
                            rss_item_title=feed_entry_title,
                            rss_item_link=feed_entry_link,
                            rss_item_guid=feed_entry_guid,
                        ).save()

                    discord_message = f"**{rss_feed.name}**\n{feed_entry_link}"

                    send_message(
                        channel_id=rss_feed.discord_channel.channel,
                        message=discord_message,
                    )
                else:
                    logger.debug(
                        msg=(
                            f'No item for feed "{rss_feed.name}" to post. '
                            'Missing either "post_entry" to be "True" or '
                            'either "feed_entry_link" or "feed_entry_guid" is "None".'
                        )
                    )
        else:
            logger.debug(msg="No RSS feeds found to parse.")
    else:
        logging.info(
            msg=(
                "AA Discordbot (https://github.com/pvyParts/allianceauth-discordbot) "
                "needs to be installed and configured."
            )
        )



"""
Managers for our models
"""

# Django
from django.db import models


class RssFeedsManager(models.Manager):
    """
    AFatManager
    """

    def select_enabled(self):
        """
        Apply select_related for default query optimizations.
        """

        return self.filter(enabled=True)
